﻿using HeuristicLab.Algorithms.PushGP.Interpreter;
using HeuristicLab.Algorithms.PushGP.Stack;

namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    /// <summary>
    /// Pushes a copy of an indexed item "deep" in the stack onto the top of the stack, without removing the deep item.
    /// The index is taken from the INTEGER stack, and the indexing is done after the index is removed.
    /// </summary>
    /// <typeparam name="T">Stacktype</typeparam>
    public abstract class YankDuplicateExpression<T> : StatelessExpression
    {
        public void Eval(IStack<T> stack, IStack<long> integerStack, int count = 1)
        {
            if (integerStack.Count == 0)
                return;

            var index = integerStack.Top;

            if (index > stack.Count - count ||
                index < 0 ||
                stack.Count < 2)
                return;

            integerStack.Pop();
            var item = stack.ElementAt((int)index);

            stack.Push(item);
        }
    }

    public class IntegerYankDuplicateExpression : YankDuplicateExpression<long>
    {
        protected override string InitStringRepresentation() { return "INTEGER.YANKDUP"; }

        public override void Eval(IInterpreter interpreter)
        {
            Eval(interpreter.IntegerStack, interpreter.IntegerStack, 2);
        }
    }

    public class FloatYankDuplicateExpression : YankDuplicateExpression<double>
    {
        protected override string InitStringRepresentation() { return "FLOAT.YANKDUP"; }

        public override void Eval(IInterpreter interpreter)
        {
            Eval(interpreter.FloatStack, interpreter.IntegerStack);
        }
    }

    public class BooleanYankDuplicateExpression : YankDuplicateExpression<bool>
    {
        protected override string InitStringRepresentation() { return "BOOLEAN.YANKDUP"; }

        public override void Eval(IInterpreter interpreter)
        {
            Eval(interpreter.BooleanStack, interpreter.IntegerStack);
        }
    }

    public class NameYankDuplicateExpression : YankDuplicateExpression<string>
    {
        protected override string InitStringRepresentation() { return "NAME.YANKDUP"; }

        public override void Eval(IInterpreter interpreter)
        {
            Eval(interpreter.NameStack, interpreter.IntegerStack);
        }
    }

    public class ExecYankDuplicateExpression : YankDuplicateExpression<Expression>
    {
        protected override string InitStringRepresentation() { return "EXEC.YANKDUP"; }

        public override void Eval(IInterpreter interpreter)
        {
            Eval(interpreter.ExecStack, interpreter.IntegerStack);
        }
    }

    public class CodeYankDuplicateExpression : YankDuplicateExpression<Expression>
    {
        protected override string InitStringRepresentation() { return "CODE.YANKDUP"; }

        public override void Eval(IInterpreter interpreter)
        {
            Eval(interpreter.CodeStack, interpreter.IntegerStack);
        }
    }
}
